/*
 * Copyright Bigeon Emmanuel (2014)
 *
 * emmanuel@bigeon.fr
 *
 * This software is a computer program whose purpose is to
 * provide a generic framework for console applications.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
/** gclc:fr.bigeon.gclc.ConsoleApplication.java
 * Created on: Sep 6, 2014 */
package fr.bigeon.gclc;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import fr.bigeon.gclc.command.HelpExecutor;
import fr.bigeon.gclc.command.ICommand;
import fr.bigeon.gclc.command.ICommandProvider;
import fr.bigeon.gclc.command.SubedCommand;
import fr.bigeon.gclc.command.UnrecognizedCommand;
import fr.bigeon.gclc.exception.CommandParsingException;
import fr.bigeon.gclc.exception.CommandRunException;
import fr.bigeon.gclc.exception.InvalidCommandName;
import fr.bigeon.gclc.i18n.Messages;
import fr.bigeon.gclc.manager.ConsoleManager;
import fr.bigeon.gclc.manager.SystemConsoleManager;
import fr.bigeon.gclc.prompt.CLIPrompterMessages;

/** <p>
 * A {@link ConsoleApplication} is an application that require the user to input
 * commands.
 * <p>
 * A typical use case is the following:
 *
 * <pre>
 * {@link ConsoleApplication} app = new {@link ConsoleApplication#ConsoleApplication(String, String, String) ConsoleApplication("exit", "welcome", "see you latter")};
 * app.{@link ConsoleApplication#add(ICommand) add}("my_command", new {@link ICommand MyCommand()});
 * app.start();
 * </pre>
 * <p>
 * That will launch in the console application that will display "welcome",
 * return a line and prompt the user for the command until he inputs the
 * <code>exit</code> command. Then it will print "see you latter" and exit the
 * start method.
 *
 * @author Emmanuel BIGEON */
public class ConsoleApplication implements ICommandProvider {

    /** The class logger */
    private static final Logger LOGGER = Logger
            .getLogger(ConsoleApplication.class.getName());
    /** The welcome message */
    private final String header;
    /** The good bye message */
    private final String footer;
    /** The console manager */
    protected final ConsoleManager manager;
    /** The container of commands */
    private final SubedCommand root;
    /** The state of this application */
    private boolean running;
    /** The listeners */
    private final List<CommandRequestListener> listeners = new ArrayList<>();

    /** @param manager the manager
     * @param welcome the welcoming message
     * @param goodbye the goodbye message */
    public ConsoleApplication(ConsoleManager manager, String welcome,
            String goodbye) {
        this.header = welcome;
        this.footer = goodbye;
        this.manager = manager;
        root = new SubedCommand(new String(), new UnrecognizedCommand(manager));
    }

    /** @param manager the manager
     * @param exit the keyword for the exit command of this application
     * @param welcome the header message to display on launch of this
     *            application
     * @param goodbye the message to display on exit
     * @throws InvalidCommandName if the exit command name is invalid */
    public ConsoleApplication(ConsoleManager manager, String exit,
            String welcome, String goodbye) throws InvalidCommandName {
        this(manager, welcome, goodbye);
        root.add(new ExitCommand(exit, this));
    }

    /** @param exit the keyword for the exit command of this application
     * @param welcome the header message to display on launch of this
     *            application
     * @param goodbye the message to display on exit
     * @throws InvalidCommandName if the exit command name is invalid */
    public ConsoleApplication(String exit, String welcome,
            String goodbye) throws InvalidCommandName {
        this(new SystemConsoleManager(), welcome, goodbye);
        root.add(new ExitCommand(exit, this));
    }

    @Override
    public final boolean add(ICommand cmd) throws InvalidCommandName {
        return root.add(cmd);
    }

    /** Adds help command on the given key
     *
     * @param cmd the handle for help
     * @return if the help command was added
     * @throws InvalidCommandName if the help command was not valid */
    public final boolean addHelpCommand(String cmd) throws InvalidCommandName {
        return root.add(new HelpExecutor(cmd, manager, root));
    }

    /** @param listener the listener to remove. */
    public final void addListener(CommandRequestListener listener) {
        listeners.add(listener);
    }

    /* (non-Javadoc)
     * @see
     * fr.bigeon.gclc.command.ICommandProvider#executeSub(java.lang.String,
     * java.lang.String[]) */
    @Override
    public final void executeSub(String command, String... args) {
        root.executeSub(command, args);
    }

    /** Exit this running application before next command prompt */
    public final void exit() {
        running = false;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.command.ICommandProvider#get(java.lang.String) */
    @Override
    public final ICommand get(String command) {
        return root.get(command);
    }

    /** @return the manager */
    public final ConsoleManager getManager() {
        return manager;
    }

    /** @param cmd the command to interpret */
    public final void interpretCommand(String cmd) {
        List<String> args;
        try {
            args = splitCommand(cmd);
        } catch (CommandParsingException e1) {
            manager.println("Command line cannot be parsed"); //$NON-NLS-1$
            LOGGER.log(Level.INFO, "Invalid user command " + cmd, e1); //$NON-NLS-1$
            return;
        }
        if (!args.isEmpty()) {
            try {
                executeSub(args.get(0), Arrays.copyOfRange(
                        args.toArray(new String[0]), 1, args.size()));
            } catch (final CommandRunException e) {
                LOGGER.log(Level.WARNING, "Command failed: " + cmd, e); //$NON-NLS-1$
                manager.println(Messages
                        .getString("ConsoleApplication.cmd.failed", cmd)); //$NON-NLS-1$
                manager.println(e.getLocalizedMessage());
            }
        }
    }

    /** Splits a command in the diferrent arguments
     * 
     * @param cmd the command to split in its parts
     * @return the list of argument preceded by the command name
     * @throws CommandParsingException if the parsing of the command failed */
    private static List<String> splitCommand(String cmd) throws CommandParsingException {
        final List<String> args = new ArrayList<>();
        // parse the string to separate arguments
        int index = 0;
        int startIndex = 0;
        boolean escaped = false;
        boolean inString = false;
        while (index < cmd.length()) {
            char c = cmd.charAt(index);
            index++;
            if (escaped || c == '\\') {
                escaped = !escaped;
                continue;
            }
            if (c == ' ' && !inString) {
                final String arg = cmd.substring(startIndex, index - 1);
                if (!arg.isEmpty()) {
                    args.add(arg);
                }
                startIndex = index;
            } else if (c == '"') {
                if (inString) {
                    inString = false;
                    args.add(endOfString(cmd, startIndex, index));
                    index++;
                    startIndex = index;
                }
                inString = startIndex == index - 1;
            }
        }
        final String arg = cmd.substring(startIndex, cmd.length());
        if (!arg.isEmpty()) {
            args.add(arg);
        }
        return args;
    }

    /** @param cmd the command to parse
     * @param startIndex the starting point of the parsing
     * @param index the index of the current position
     * @return the argument
     * @throws CommandParsingException if the end of string does not mark end of
     *             command and is not followed by a space */
    private static String endOfString(String cmd, int startIndex,
                                      int index) throws CommandParsingException {
        if (index + 1 < cmd.length() && cmd.charAt(index + 1) != ' ') {
            throw new CommandParsingException("Misplaced quote"); //$NON-NLS-1$
        }
        return cmd.substring(startIndex + 1, index - 1);
    }

    /** @param listener the listener to remove */
    public final void removeListener(CommandRequestListener listener) {
        for (int i = 0; i < listeners.size(); i++) {
            if (listeners.get(i) == listener) {
                listeners.remove(i);
                return;
            }
        }
    }

    /** Launches the prompting application */
    public final void start() {
        if (header != null) {
            manager.println(header);
        }
        running = true;
        do {
            final String cmd = manager.prompt();
            if (cmd.isEmpty()) {
                continue;
            }
            for (final CommandRequestListener listener : listeners) {
                listener.commandRequest(cmd);
            }
            interpretCommand(cmd);
        } while (running);
        if (footer != null) {
            manager.println(footer);
        }
    }
}

/** <p>
 * A command to exit a {@link ConsoleApplication}.
 *
 * @author Emmanuel BIGEON */
class ExitCommand implements ICommand {
    /** The exit command manual message key */
    private static final String EXIT_MAN = "exit.man"; //$NON-NLS-1$
    /** The tip of the exit command */
    private static final String EXIT = "exit.tip"; //$NON-NLS-1$
    /** The application that will be exited when this command runs */
    private final ConsoleApplication app;
    /** The exit command name */
    private final String name;

    /** @param name the name of the command
     * @param app the application to exit */
    public ExitCommand(String name, ConsoleApplication app) {
        this.name = name;
        this.app = app;
    }

    /** The actions to take before exiting */
    public void beforeExit() {
        // Do nothing by default
    }

    @Override
    public final void execute(String... args) {
        beforeExit();
        app.exit();
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.command.ICommand#getCommandName() */
    @Override
    public String getCommandName() {
        return name;
    }

    @Override
    public void help(ConsoleManager manager, String... args) {
        manager.println(
                CLIPrompterMessages.getString(EXIT_MAN, (Object[]) args));
    }

    @Override
    public String tip() {
        return CLIPrompterMessages.getString(EXIT);
    }
}
